//------------------------------------------------------------------------
// Board Library : Seeed nRF52 Borads 1.1.8
// Board Select  : Seeed XIAO nRF52840 Sense
// 2024/07/27 @msfujino
//------------------------------------------------------------------------

#include <SPI.h>
#include <SD.h>

#include <PDM.h>        // \Arduino15\packages\Seeeduino\hardware\nrf52\1.1.8\libraries\PDM\src
#include <RingBuf.h>    // RingBuffer by Jean-Lue - Locoduino 1.0.5
#include <FIR.h>        // FIR filter by Leeman Geophysical LLC 0.1.1

#define FIFO_SIZE   1024        // short(16 bits) max fifo size is 692 in my case
#define BUFF_SIZE   512         // short(16 bits)
#define SW          D1          // Record end switch
#define SD_CS       D2          // SD CS
#define TEST1       D3          // for digital scope
#define TEST2       D6          // for digital scope
#define logFileName "LOG.wav"   // log file name

File logFile;                   // File instance
RingBuf<short, FIFO_SIZE> fifo; // fifo instance
FIR<float, 32> fir_lp;          // FIR instance
// FIR filter designed with http://t-filter.engineerjs.com/
// sampling frequency: 16000 Hz, fixed point precision: 16 bits
// 0 Hz - 1.5 kHz : gain = 1, 2 kHz - 8 kHz : gain = 0, desired attenuation = -40 dB
static float filter_taps[32] = {419,637,900,1010,851,379,-342,-1125,
                                -1699,-1784,-1191,98,1897,3841,5485,6425,
                                6425,5485,3841,1897,98,-1191,-1784,-1699,
                                -1125,-342,379,851,1010,900,637,419};
short sampleBuffer[BUFF_SIZE];  // short(16 bits)
volatile int samplesRead;       // number of samples
int maxfifosize = 0;            // for debug

// *************************************************************** setup()
void setup() {
  Serial.begin(115200);
  while(!Serial);
  delay(2000);

  pinMode(LED_RED, OUTPUT);     // LED LOW:ON, HIGH:OFF
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);
  pinMode(TEST1, OUTPUT);       // for digital scope
  pinMode(TEST2, OUTPUT);
  pinMode(SW, INPUT_PULLUP);    // recording end switch

  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);
  digitalWrite(TEST1, LOW);
  digitalWrite(TEST2, LOW);

  if(!SD.begin(32000000UL, SD_CS)) {            // SPI clock 32MHz
    Serial.println("Failed to initialize SD!");
    errorBlink(2);
  } 
  logFile = SD.open(logFileName, FILE_WRITE);
  if(!logFile) { 
    Serial.println("Failed to open file");
    errorBlink(3);
  }
  create_template(logFile);       // wave file headder   

  PDM.onReceive(onPDMdata);       // configure the data receive callback
  PDM.setBufferSize(BUFF_SIZE);

  if (!PDM.begin(1, 16000)) {     // mono 16000 samples/sec
    Serial.println("Failed to start PDM!");
    PDM.end();
    errorBlink(4);
  }
  PDM.setGain(70);                // 0 - 80, Note:set above 70, distorted.

  // Set the FIR LPF coefficients
  fir_lp.setFilterCoeffs(filter_taps);
  // Set the gain
  Serial.print("Low Pass Filter Gain: ");
  Serial.println(fir_lp.getGain());
  
  Serial.println("end of initialization");
}

// ************************************************************** loop()
void loop() 
{
  if (samplesRead) {                      // wait for samples to be read
    digitalWrite(LED_GREEN, LOW);

    short temp;
    while(!fifo.isEmpty()) {
      digitalWrite(TEST1, HIGH);
      fifo.lockedPop(temp);               // pop from fifo
      fifo.lockedPop(temp);               // 1/4 down sampling
      fifo.lockedPop(temp);
      fifo.lockedPop(temp);
      logFile.write(byte(temp & 0x00FF)); // little-endian
      logFile.write(byte(temp >> 8));
    }
//    digitalWrite(TEST1, HIGH);
    digitalWrite(TEST1, LOW);             // in my case 660us every 16mS, sometimes 37mS   
    samplesRead = 0;

    digitalWrite(LED_GREEN, HIGH);
  }
  
  if(digitalRead(SW) == LOW) {            // recording end ?
    PDM.end();
    Serial.println("Finished recording");

    logFile.close();
    logFile = SD.open(logFileName, O_WRITE);
    finalize_template(logFile);           // finalize
    Serial.print("max fifo size\t"); Serial.println(maxfifosize);
    errorBlink(10);
  }
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ onPDMdata()
void onPDMdata() 
{
  digitalWrite(TEST2, HIGH);
  digitalWrite(LED_BLUE, LOW);

  int bytesAvailable = PDM.available();     // the number of bytes available
  PDM.read(sampleBuffer, bytesAvailable);   // read into the sample buffer
  samplesRead = bytesAvailable / 2;         // 16-bit, 2 bytes per sample

  for(int i = 0; i < samplesRead; i++) {    // push to fifo
    // filtering
  //  digitalWrite(TEST2, HIGH);
    short filterData = fir_lp.processReading(sampleBuffer[i]);
  //  digitalWrite(TEST2, LOW);             // filter excution time 9.4uS
    if(!fifo.lockedPush(filterData)) {
      PDM.end();
      Serial.print("maxfifosize "); Serial.print(maxfifosize);
      logFile.close();
      logFile = SD.open(logFileName, O_WRITE);
      finalize_template(logFile);           // finalize
      errorBlink(5);
    }
    // to set the FIFO size, find the maximum FIFO size used
    int fifosize = fifo.size();
    if(maxfifosize < fifosize) { maxfifosize = fifosize; }
  }

  digitalWrite(LED_BLUE, HIGH);
  digitalWrite(TEST2, LOW);                 // in nmy case 2.5mS every 16mS
}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// See \Arduino15\packages\Seeeduino\hardware\mbed\2.9.2\libraries\PDM\examples\PDMSavedOnSDcard\PDMSavedOnSDcard.ino
void create_template(File &sFile)
{    
  struct soundhdr {
    char  riff[4];        // "RIFF"                                  
    long  flength;        // file length in bytes                    
    char  wave[4];        // "WAVE"                                  
    char  fmt[4];         // "fmt "                                  
    long  chunk_size;     // size of FMT chunk in bytes (usually 16) 
    short format_tag;     // 1=PCM, 257=Mu-Law, 258=A-Law, 259=ADPCM 
    short num_chans;      // 1=mono, 2=stereo                        
    long  srate;          // Sampling rate in samples per second     
    long  bytes_per_sec;  // bytes per second = srate*bytes_per_samp 
    short bytes_per_samp; // 2=16-bit mono, 4=16-bit stereo          
    short bits_per_samp;  // Number of bits per sample               
    char  data[4];        // "data"                                  
    long  dlength;        // data length in bytes (filelength - 44)  
  } wavh;

  // It's easy enough to initialize the strings
  strncpy(wavh.riff,"RIFF", 4);
  strncpy(wavh.wave,"WAVE", 4);
  strncpy(wavh.fmt,"fmt ", 4);
  strncpy(wavh.data,"data", 4);
  
  // size of FMT chunk in bytes
  wavh.chunk_size = 16;
  wavh.format_tag = 1;  // PCM
  wavh.num_chans = 1;   // mono
//  wavh.srate = 16000;
//  wavh.bytes_per_sec = (16000 * 1 * 16 * 1)/8;
  wavh.srate = 4000;
  wavh.bytes_per_sec = (4000 * 1 * 16 * 1)/8;
  wavh.bytes_per_samp = 2;
  wavh.bits_per_samp = 16;

  sFile.seek(0);
  sFile.write((byte *)&wavh, 44);
}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void finalize_template(File &sFile)
{
  unsigned long fSize = sFile.size();
  unsigned long ftemp = fSize - 8;
  unsigned long dtemp = fSize - 36;
  Serial.print("file size\t"); Serial.print(fSize); Serial.print("\t0x"); Serial.println(fSize, HEX);
  Serial.print("flength\t\t"); Serial.print(ftemp); Serial.print("\t0x"); Serial.println(ftemp, HEX);
  Serial.print("dlength\t\t"); Serial.print(dtemp); Serial.print("\t0x"); Serial.println(dtemp, HEX);

  sFile.seek(4);
  byte fdata[4] = {lowByte(ftemp), highByte(ftemp), ftemp >> 16, ftemp >> 24};
  sFile.write((byte*)fdata, 4);

  sFile.seek(40);
  byte ddata[4] = {lowByte(dtemp), highByte(dtemp), dtemp >> 16, dtemp >> 24}; 
  sFile.write((byte*)ddata, 4);

  sFile.close();
}

//********************************************************************* errorBlink()
// Blinks specified number of times on error
// used delayMicroseconds() insted of delay()
void errorBlink(uint8_t err)
{  
  Serial.print("ERROR "); Serial.println(err); Serial.flush();
  while(true) {      
    digitalWrite(LED_GREEN, HIGH);
    digitalWrite(LED_BLUE, HIGH);
    for(int i = 0; i < err; i++) {
      digitalWrite(LED_RED, LOW);
      delayMicroseconds(50000);
      digitalWrite(LED_RED, HIGH);
      delayMicroseconds(500000);
    }
  delayMicroseconds(1000000);
  }
}